"""
sliderule
=========

A C++/Lua framework for on-demand science data processing
"""
from .sliderule import *
import sliderule.version

# get semantic version from setuptools-scm
__version__ = sliderule.version.version
